@@ -1,9 +1,9 @@
-/* Copyright 2000-2005 The Apache Software Foundation or its licensors, as
- * applicable.
- *
- * Licensed under the Apache License, Version 2.0 (the "License");
- * you may not use this file except in compliance with the License.
- * You may obtain a copy of the License at
+/* Licensed to the Apache Software Foundation (ASF) under one or more
+ * contributor license agreements.  See the NOTICE file distributed with
+ * this work for additional information regarding copyright ownership.
+ * The ASF licenses this file to You under the Apache License, Version 2.0
+ * (the "License"); you may not use this file except in compliance with
+ * the License.  You may obtain a copy of the License at
  *
  *     http://www.apache.org/licenses/LICENSE-2.0
  *
@@ -126,14 +126,17 @@ typedef void*                 apr_os_shm_t;
 /** Basic OS process mutex structure. */
 struct apr_os_proc_mutex_t {
 #if APR_HAS_SYSVSEM_SERIALIZE || APR_HAS_FCNTL_SERIALIZE || APR_HAS_FLOCK_SERIALIZE
+    /** Value used for SYS V Semaphore, FCNTL and FLOCK serialization */
     int crossproc;
 #endif
 #if APR_HAS_PROC_PTHREAD_SERIALIZE
+    /** Value used for PTHREAD serialization */
     pthread_mutex_t *pthread_interproc;
 #endif
 #if APR_HAS_THREADS
     /* If no threads, no need for thread locks */
 #if APR_USE_PTHREAD_SERIALIZE
+    /** This value is currently unused within APR and Apache */ 
     pthread_mutex_t *intraproc;
 #endif
 #endif
@@ -183,9 +186,7 @@ struct apr_os_sock_info_t {
     struct sockaddr *remote; /**< NULL if not connected */
     int family;             /**< always required (APR_INET, APR_INET6, etc.) */
     int type;               /**< always required (SOCK_STREAM, SOCK_DGRAM, etc.) */
-#ifdef APR_ENABLE_FOR_1_0   /**< enable with APR 1.0 */
     int protocol;           /**< 0 or actual protocol (APR_PROTO_SCTP, APR_PROTO_TCP, etc.) */
-#endif
 };
 
 typedef struct apr_os_sock_info_t apr_os_sock_info_t;
